@@ -8,15 +8,16 @@
 Mboard implementation base class
 """
 
-from __future__ import print_function
 import os
+from enum import Enum
 from hashlib import md5
 from time import sleep
 from concurrent import futures
-from builtins import str
-from builtins import object
 from six import iteritems, itervalues
 from usrp_mpm.mpmlog import get_logger
+from usrp_mpm.sys_utils.filesystem_status import get_fs_version
+from usrp_mpm.sys_utils.filesystem_status import get_mender_artifact
+from usrp_mpm.sys_utils.udev import get_eeprom_paths_by_symbol
 from usrp_mpm.sys_utils.udev import get_eeprom_paths
 from usrp_mpm.sys_utils.udev import get_spidev_nodes
 from usrp_mpm.sys_utils import dtoverlay
@@ -47,12 +48,18 @@ def get_dboard_class_from_pid(pid):
 # pylint: disable=no-self-use
 # pylint: disable=too-many-public-methods
 # pylint: disable=too-many-instance-attributes
-class PeriphManagerBase(object):
+class PeriphManagerBase:
     """"
     Base class for all motherboards. Common function and API calls should
     be implemented here. Motherboard specific information can be stored in
     separate motherboard classes derived from this class
     """
+    class _EepromSearch(Enum):
+        """
+        List supported ways of searching EEPROM files.
+        """
+        LEGACY = 1 # Using EEPROM address
+        SYMBOL = 2 # Using symbol names
     #########################################################################
     # Overridables
     #
@@ -64,6 +71,9 @@ class PeriphManagerBase(object):
     pids = {}
     # A textual description of this device type
     description = "MPM Device"
+    # EEPROM layout used by this class. Defaults to legacy which uses eeprom.py
+    # to read EEPROM data
+    eeprom_search = _EepromSearch.LEGACY
     # Address of the motherboard EEPROM. This could be something like
     # "e0005000.i2c". This value will be passed to get_eeprom_paths() tos
     # determine a full path to an EEPROM device.
@@ -105,6 +115,10 @@ class PeriphManagerBase(object):
     # The index of the first port of the RFNoC crossbar which is connected to
     # an RFNoC block
     crossbar_base_port = 0
+    # A DboardIface class which will be passed to the discovered DB
+    # constructors.
+    # If None, the MB does not support the DB Iface architecture.
+    db_iface = None
     # Address of the daughterboard EEPROMs. This could be something like
     # "e0004000.i2c". This value will be passed to get_eeprom_paths() to
     # determine a full path to an EEPROM device.
@@ -145,6 +159,18 @@ class PeriphManagerBase(object):
     # a claim(). Override and set to True in the derived class if desired.
     clear_rpc_registry_on_unclaim = False
 
+    # Symbols are use to find board EEPROM by a symbolic name. Can only be used
+    # on systems that support symbol name under /proc/device-tree/__symbols__.
+    # symbol name for motherboard EEPROM file
+    mboard_eeprom_symbol = "mb_eeprom"
+    # symbol glob for daugtherboard EEPROM files
+    dboard_eeprom_symbols = "db[0,1]_eeprom"
+    # symbol glob fox auxiliary boards
+    auxboard_eeprom_symbols = "*aux_eeprom"
+    # List of discoverable features supported by a device.
+    discoverable_features = []
+
+
     # Disable checks for unused args in the overridables, because the default
     # implementations don't need to use them.
     # pylint: disable=unused-argument
@@ -166,19 +192,20 @@ class PeriphManagerBase(object):
             version_string = ""
         mboard_info["mpm_sw_version"] = version_string
 
+        fs_version = get_fs_version()
+        if fs_version is not None:
+            mboard_info["fs_version"] = fs_version
+        # Mender artifacts are generally not present on a machine hosting
+        # a simulated device--let it slide if not found on sim devices
         try:
-            with open("/etc/version", "r") as version_file:
-                mboard_info["fs_version"] = version_file.read().strip(" \r\n")
+            mboard_info['mender_artifact'] = get_mender_artifact()
         except FileNotFoundError:
-            mboard_info["fs_version"] = "FILE NOT FOUND"
-
-        try:
-            with open("/etc/mender/artifact_info", "r") as artifact_file:
-                for line in artifact_file.read().splitlines():
-                    if line.startswith('artifact_name='):
-                        mboard_info['mender_artifact'] = line[14:]
-        except FileNotFoundError:
-            mboard_info['mender_artifact'] = "FILE NOT FOUND"
+            # Note that the simulated key will not be present for
+            # non-simulated devices, hence the use of get()
+            if mboard_info.get('simulated', '') == 'True':
+                pass
+            else:
+                raise
 
         for i,dboard_info in enumerate(dboard_infos):
             mboard_info["dboard_{}_pid".format(i)] = str(dboard_info["pid"])
@@ -213,22 +240,24 @@ class PeriphManagerBase(object):
         assert self.mboard_eeprom_magic is not None
         self.dboards = []
         self._default_args = ""
+        # CHDR transport managers. These need to be instantiated by the child
+        # classes.
+        self._xport_mgrs = {}
         # Set up logging
         self.log = get_logger('PeriphManager')
         self.claimed = False
         try:
-            self._eeprom_head, self._eeprom_rawdata = \
-                self._read_mboard_eeprom()
-            self.mboard_info = self._get_mboard_info(self._eeprom_head)
+            self.mboard_info = self._get_mboard_info()
             self.log.info("Device serial number: {}"
                           .format(self.mboard_info.get('serial', 'n/a')))
-            self.dboard_infos = self._get_dboard_eeprom_info()
+            self.dboard_infos = self._get_dboard_info()
             self.device_info = \
                     self.generate_device_info(
                         self._eeprom_head,
                         self.mboard_info,
                         self.dboard_infos
                     )
+            self._aux_board_infos = self._get_aux_board_info()
         except BaseException as ex:
             self.log.error("Failed to initialize device: %s", str(ex))
             self._device_initialized = False
@@ -264,63 +293,120 @@ class PeriphManagerBase(object):
         self._device_initialized = True
         self._initialization_status = "No errors."
 
-    def _read_mboard_eeprom(self):
-        """
-        Read out mboard EEPROM.
-        Returns a tuple: (eeprom_dict, eeprom_rawdata), where the the former is
-        a de-serialized dictionary representation of the data, and the latter
-        is a binary string with the raw data.
-
-        If no EEPROM is defined, returns empty values.
-        """
-        if self.mboard_eeprom_addr:
-            self.log.trace("Reading EEPROM from address `{}'..."
-                           .format(self.mboard_eeprom_addr))
-            eeprom_paths = get_eeprom_paths(self.mboard_eeprom_addr)
-            if not eeprom_paths:
-                self.log.error("Could not identify EEPROM paths for %s!",
-                               self.mboard_eeprom_addr)
-                return {}, b''
-            self.log.trace("Found mboard EEPROM path: %s", eeprom_paths[0])
-            (eeprom_head, eeprom_rawdata) = eeprom.read_eeprom(
-                eeprom_paths[0],
+    def _read_mboard_eeprom_data(self, path):
+        return eeprom.read_eeprom(
+                path,
                 self.mboard_eeprom_offset,
                 eeprom.MboardEEPROM.eeprom_header_format,
                 eeprom.MboardEEPROM.eeprom_header_keys,
                 self.mboard_eeprom_magic,
-                self.mboard_eeprom_max_len,
-            )
-            self.log.trace("Found EEPROM metadata: `{}'"
-                           .format(str(eeprom_head)))
-            self.log.trace("Read {} bytes of EEPROM data."
-                           .format(len(eeprom_rawdata)))
-            return eeprom_head, eeprom_rawdata
-        # Nothing defined? Return defaults.
-        self.log.trace("No mboard EEPROM path defined. "
-                       "Skipping mboard EEPROM readout.")
-        return {}, b''
+                self.mboard_eeprom_max_len)
 
-    def _get_mboard_info(self, eeprom_head):
+    def _read_mboard_eeprom_legacy(self):
+        """
+        Read out mboard EEPROM.
+        Saves _eeprom_head, _eeprom_rawdata as class members where the the
+        former is a de-serialized dictionary representation of the data, and the
+        latter is a binary string with the raw data.
+
+        If no EEPROM both members are defined and empty.
+        """
+        if not self.mboard_eeprom_addr:
+            self.log.trace("No mboard EEPROM path defined. "
+                           "Skipping mboard EEPROM readout.")
+            return
+
+        self.log.trace("Reading EEPROM from address `{}'..."
+                       .format(self.mboard_eeprom_addr))
+        eeprom_paths = get_eeprom_paths(self.mboard_eeprom_addr)
+        if not eeprom_paths:
+            self.log.error("Could not identify EEPROM paths for %s!",
+                           self.mboard_eeprom_addr)
+            return
+
+        self.log.trace("Found mboard EEPROM path: %s", eeprom_paths[0])
+        (self._eeprom_head, self._eeprom_rawdata) = \
+            self._read_mboard_eeprom_data(eeprom_paths[0])
+
+    def _read_mboard_eeprom_by_symbol(self):
+        """
+        Read out mboard EEPROM.
+        Saves _eeprom_head, _eeprom_rawdata as class members where the the
+        former is a de-serialized dictionary representation of the data, and the
+        latter is a binary string with the raw data.
+
+        If no EEPROM both members are defined and empty.
+        """
+        if not self.mboard_eeprom_symbol:
+            self.log.trace("No mboard EEPROM path defined. "
+                           "Skipping mboard EEPROM readout.")
+            return
+
+        self.log.trace("Reading EEPROM from address `{}'..."
+                       .format(self.mboard_eeprom_symbol))
+        eeprom_paths = get_eeprom_paths_by_symbol(self.mboard_eeprom_symbol)
+        if not eeprom_paths:
+            self.log.error("Could not identify EEPROM paths for %s!",
+                           self.mboard_eeprom_addr)
+            return
+        # There should be exact one item in the dictionary returned by
+        # find_eeprom_paths. If so take the value of it as path for further
+        # processing
+        if not (len(eeprom_paths) == 1):
+            raise RuntimeError("System should contain exact one EEPROM file"
+                               "for motherboard but found %d." % len(eeprom_paths))
+        eeprom_path = str(eeprom_paths.popitem()[1])
+
+        self.log.trace("Found mboard EEPROM path: %s", eeprom_path)
+        (self._eeprom_head, self._eeprom_rawdata) = \
+            self._read_mboard_eeprom_data(eeprom_path)
+
+    def _read_mboard_eeprom(self):
+        """
+        Read out mboard EEPROM.
+
+        This is a wrapper call to switch between the support EEPROM layouts.
+        """
+        if not self.eeprom_search in self._EepromSearch:
+            self.log.warning("%s is not a valid EEPROM layout type. "
+                             "Skipping readout.")
+            return
+
+        self._eeprom_head, self._eeprom_rawdata = {}, b""
+        if self.eeprom_search == self._EepromSearch.LEGACY:
+            self._read_mboard_eeprom_legacy()
+        elif self.eeprom_search == self._EepromSearch.SYMBOL:
+            self._read_mboard_eeprom_by_symbol()
+
+        self.log.trace("Found EEPROM metadata: `{}'"
+                       .format(str(self._eeprom_head)))
+        self.log.trace("Read {} bytes of EEPROM data."
+                       .format(len(self._eeprom_rawdata)))
+
+    def _get_mboard_info(self):
         """
         Creates the mboard info dictionary from the EEPROM data.
         """
+        if not hasattr(self, "_eeprom_head"):
+            # read eeprom if not done already
+            self._read_mboard_eeprom()
         mboard_info = self.mboard_info
-        if not eeprom_head:
+        if not self._eeprom_head:
             self.log.debug("No EEPROM info: Can't generate mboard_info")
             return mboard_info
         for key in ('pid', 'serial', 'rev', 'eeprom_version'):
             # In C++, we can only handle dicts if all the values are of the
             # same type. So we must convert them all to strings here:
             try:
-                mboard_info[key] = str(eeprom_head.get(key, ''), 'ascii')
+                mboard_info[key] = str(self._eeprom_head.get(key, ''), 'ascii')
             except TypeError:
-                mboard_info[key] = str(eeprom_head.get(key, ''))
-        if 'pid' in eeprom_head:
-            if eeprom_head['pid'] not in self.pids.keys():
+                mboard_info[key] = str(self._eeprom_head.get(key, ''))
+        if 'pid' in self._eeprom_head:
+            if self._eeprom_head['pid'] not in self.pids.keys():
                 self.log.error(
                     "Found invalid PID in EEPROM: 0x{:04X}. " \
                     "Valid PIDs are: {}".format(
-                        eeprom_head['pid'],
+                        self._eeprom_head['pid'],
                         ", ".join(["0x{:04X}".format(x) for x in self.pids]),
                     )
                 )
@@ -329,7 +415,7 @@ class PeriphManagerBase(object):
         # back to the the rev itself (because every rev is compatible with
         # itself).
         rev_compat = \
-            eeprom_head.get('rev_compat', eeprom_head.get('rev'))
+            self._eeprom_head.get('rev_compat', self._eeprom_head.get('rev'))
         try:
             rev_compat = int(rev_compat)
         except (ValueError, TypeError):
@@ -353,7 +439,16 @@ class PeriphManagerBase(object):
                 )
         return mboard_info
 
-    def _get_dboard_eeprom_info(self):
+    def _read_dboard_eeprom_data(self, path):
+        return eeprom.read_eeprom(
+            path,
+            self.dboard_eeprom_offset,
+            eeprom.DboardEEPROM.eeprom_header_format,
+            eeprom.DboardEEPROM.eeprom_header_keys,
+            self.dboard_eeprom_magic,
+            self.dboard_eeprom_max_len)
+
+    def _get_dboard_info_legacy(self):
         """
         Read back EEPROM info from the daughterboards
         """
@@ -378,14 +473,8 @@ class PeriphManagerBase(object):
         dboard_info = []
         for dboard_idx, dboard_eeprom_path in enumerate(dboard_eeprom_paths):
             self.log.debug("Reading EEPROM info for dboard %d...", dboard_idx)
-            dboard_eeprom_md, dboard_eeprom_rawdata = eeprom.read_eeprom(
-                dboard_eeprom_path,
-                self.dboard_eeprom_offset,
-                eeprom.DboardEEPROM.eeprom_header_format,
-                eeprom.DboardEEPROM.eeprom_header_keys,
-                self.dboard_eeprom_magic,
-                self.dboard_eeprom_max_len,
-            )
+            dboard_eeprom_md, dboard_eeprom_rawdata = \
+                self._read_dboard_eeprom_data(dboard_eeprom_path)
             self.log.trace("Found dboard EEPROM metadata: `{}'"
                            .format(str(dboard_eeprom_md)))
             self.log.trace("Read %d bytes of dboard EEPROM data.",
@@ -403,6 +492,95 @@ class PeriphManagerBase(object):
             })
         return dboard_info
 
+    def _get_board_info_by_symbol(self, symbols):
+        """
+        Collect board info for given symbols.
+        symbols: a (glob) expression identifying EEPROMs to search for
+        returns: dictionary of EEPROM content found with symbol name as key
+                 and an dictonary wih metadeta, rawdata and pid as value
+        """
+        result = {}
+        self.log.trace("Identifying EEPROM paths from %s...", symbols)
+        eeprom_paths = get_eeprom_paths_by_symbol(symbols)
+        self.log.trace("Found EEPROM paths: %s", eeprom_paths)
+        for name, path in eeprom_paths.items():
+            self.log.debug("Reading EEPROM info for %s...", name)
+            if not path:
+                if "db" in name:
+                    # In order to support having a single dboard in slot 1
+                    # with slot 0 empty on a x4xx, we pretend that there is
+                    # a dummy "EmptyDaughterboard" here.
+                    self.log.debug("Not present. Inserting dummy DB info")
+                    result[name] = {
+                        'eeprom_md': {'serial': 'deadbee', 'pid': 0x0},
+                        'eeprom_raw': [],
+                        'pid': 0x0
+                    }
+                else:
+                    self.log.debug("Not present. Skipping board")
+                continue
+            try:
+                eeprom_md, eeprom_rawdata = self._read_dboard_eeprom_data(path)
+                self.log.trace("Found EEPROM metadata: `{}'"
+                               .format(str(eeprom_md)))
+                self.log.trace("Read %d bytes of dboard EEPROM data.",
+                               len(eeprom_rawdata))
+                pid = eeprom_md.get('pid')
+                if pid is None:
+                    self.log.warning("No PID found in EEPROM!")
+                else:
+                    self.log.debug("Found PID in EEPROM: 0x{:04X}".format(pid))
+                result[name] = {'eeprom_md': eeprom_md,
+                                'eeprom_rawdata': eeprom_rawdata,
+                                'pid': pid}
+            except RuntimeError as e:
+                self.log.warning("Could not read EEPROM for %s (%s)", name, e)
+
+        return result
+
+    def _get_dboard_info_by_symbol(self):
+        """
+        Read back EEPROM info from the daughterboards
+        """
+        dboard_info = self._get_board_info_by_symbol(self.dboard_eeprom_symbols)
+        if len(dboard_info) > self.max_num_dboards:
+            self.log.warning("Found more EEPROM paths than daughterboards. "
+                             "Ignoring some of them.")
+            # dboard_infos keys are sorted so it is safe to remove all items
+            # but the first few. Assumption is that the board names are given
+            # sorted such as db0, db1, db2, …, dbn.
+            dboard_info = {key: val for key, val in dboard_info.items()
+                    if key in list(dboard_info.keys())[:self.max_num_dboards]}
+
+        # convert dboard dict back to list for backward compatibility
+        return [value for key, value in dboard_info.items() if value]
+
+    def _get_dboard_info(self):
+        """
+        Read back EEPROM info from the daughterboards
+        """
+        if not self.eeprom_search in self._EepromSearch:
+            self.log.warning("%s is not a valid EEPROM search type. "
+                             "Skipping readout.")
+            return []
+
+        if self.eeprom_search == self._EepromSearch.LEGACY:
+            return self._get_dboard_info_legacy()
+        if self.eeprom_search == self._EepromSearch.SYMBOL:
+            return self._get_dboard_info_by_symbol()
+
+    def _get_aux_board_info(self):
+        """
+        Read back EEPROM info from all auxiliary boards
+        """
+        if self.eeprom_search == self._EepromSearch.LEGACY:
+            #legacy has no support for aux board EEPROM read
+            return {}
+        self.log.debug("Read aux boards EEPROMs")
+        result = self._get_board_info_by_symbol(self.auxboard_eeprom_symbols)
+        self.log.trace("Found aux board info for: %s.", ", ".join(result.keys()))
+        return result
+
     def _update_default_args(self, default_args):
         """
         Pipe the default_args (that get passed into us from the RPC server)
@@ -442,7 +620,7 @@ class PeriphManagerBase(object):
         Initialize all the daughterboards
 
         dboard_infos -- List of dictionaries as returned from
-                       _get_dboard_eeprom_info()
+                       _get_dboard_info()
         override_dboard_pids -- List of dboard PIDs to force
         default_args -- Default args
         """
@@ -468,15 +646,62 @@ class PeriphManagerBase(object):
                 self.log.trace("Found spidev nodes: {0}".format(spi_nodes))
             else:
                 spi_nodes = []
-                self.log.warning("No SPI nodes for dboard %d.", dboard_idx)
+                self.log.trace("No SPI nodes for dboard %d.", dboard_idx)
             dboard_info.update({
                 'spi_nodes': spi_nodes,
                 'default_args': default_args,
             })
+            # If the MB supports the DB Iface architecture, pass
+            # the corresponding DB Iface to the dboard class
+            if self.db_iface is not None:
+                dboard_info['db_iface'] = self.db_iface(dboard_idx, self)
             # This will actually instantiate the dboard class:
             self.dboards.append(db_class(dboard_idx, **dboard_info))
         self.log.info("Initialized %d daughterboard(s).", len(self.dboards))
 
+    def _add_public_methods(self, src, prefix="", filter_cb=None, allow_overwrite=False):
+        """
+        Add public methods (=API) of src to self. To avoid naming conflicts and
+        make relations clear, all added method names are prefixed with 'prefix'.
+
+        Example:
+        >>> class Foo:
+        ...     def print_x(self, x):
+        ...         print(x)
+        ...
+        >>> foo = Foo()
+        >>> self._add_public_methods(foo, prefix="ext")
+        >>> self.ext_print_x(5) # Prints 5
+
+        :param source: The object to import the API from
+        :param prefix: method names in dest will be prefixed with prefix
+        :param filter_cb: A callback that returns true if the method should be
+                          added. Defaults to always returning True
+        :param allow_overwrite: If True, then methods from src will overwrite
+                                existing methods on self. Use with care.
+        """
+        filter_cb = filter_cb or (lambda *args: True)
+        assert callable(filter_cb)
+        self.log.trace("Adding API functions from %s to %s" % (
+            src.__class__.__name__, self.__class__.__name__))
+        # append _ to prefix if it is not an empty string
+        if prefix:
+            prefix = prefix + "_"
+        for name in [name for name in dir(src)
+                     if not name.startswith("_")
+                     and callable(getattr(src, name))
+                     and filter_cb(name, getattr(src, name))
+                     ]:
+            destname = prefix + name
+            if hasattr(self, destname) and not allow_overwrite:
+                self.log.warn("Cannot add method {} because it would "
+                              "overwrite existing method.".format(destname))
+            else:
+                method = getattr(src, name)
+                self.log.trace("Add function %s as %s", name, destname)
+                setattr(self, destname, method)
+
+
     ###########################################################################
     # Session (de-)initialization (at UHD startup)
     ###########################################################################
@@ -505,6 +730,8 @@ class PeriphManagerBase(object):
             self.log.error(
                 "Cannot run init(), device was never fully initialized!")
             return False
+        for xport_mgr in self._xport_mgrs.values():
+            xport_mgr.init(args)
         if not self.dboards:
             return True
         if args.get("serialize_init", False):
@@ -533,8 +760,11 @@ class PeriphManagerBase(object):
                 "Cannot run deinit(), device was never fully initialized!")
             return
         self.log.trace("Mboard deinit() called.")
-        for dboard in self.dboards:
+        for slot, dboard in enumerate(self.dboards):
+            self.log.trace("call deinit() on dBoard in slot {}".format(slot))
             dboard.deinit()
+        for xport_mgr in self._xport_mgrs.values():
+            xport_mgr.deinit()
 
     def tear_down(self):
         """
@@ -542,6 +772,8 @@ class PeriphManagerBase(object):
         deconstruction.
         """
         self.log.trace("Teardown called for Peripheral Manager base.")
+        for each in self.dboards:
+            each.tear_down()
 
     ###########################################################################
     # RFNoC & Device Info
@@ -685,6 +917,7 @@ class PeriphManagerBase(object):
         assert (len(metadata_l) == len(data_l)),\
             "update_component arguments must be the same length"
         # Iterate through the components, updating each in turn
+        basepath = os.path.join(os.sep, "tmp", "uploads")
         for metadata, data in zip(metadata_l, data_l):
             id_str = metadata['id']
             filename = os.path.basename(metadata['filename'])
@@ -693,7 +926,7 @@ class PeriphManagerBase(object):
                     id_str, self.updateable_components.keys()
                 ))
                 raise KeyError("Update component not implemented for {}".format(id_str))
-            self.log.trace("Updating component: {}".format(id_str))
+            self.log.trace("Downloading component: {}".format(id_str))
             if 'md5' in metadata:
                 given_hash = metadata['md5']
                 comp_hash = md5()
@@ -710,10 +943,9 @@ class PeriphManagerBase(object):
                                        comp_hash, given_hash))
                     raise RuntimeError("Component file hash mismatch")
             else:
-                self.log.trace("Loading unverified {} image.".format(
+                self.log.trace("Downloading unhashed {} image.".format(
                     id_str
                 ))
-            basepath = os.path.join(os.sep, "tmp", "uploads")
             filepath = os.path.join(basepath, filename)
             if not os.path.isdir(basepath):
                 self.log.trace("Creating directory {}".format(basepath))
@@ -721,9 +953,15 @@ class PeriphManagerBase(object):
             self.log.trace("Writing data to {}".format(filepath))
             with open(filepath, 'wb') as comp_file:
                 comp_file.write(data)
+
+        # do the actual installation on the device
+        for metadata in metadata_l:
+            id_str = metadata['id']
+            filename = os.path.basename(metadata['filename'])
+            filepath = os.path.join(basepath, filename)
             update_func = \
                 getattr(self, self.updateable_components[id_str]['callback'])
-            self.log.info("Updating component `%s'", id_str)
+            self.log.info("Installing component `%s'", id_str)
             update_func(filepath, metadata)
         return True
 
@@ -740,7 +978,7 @@ class PeriphManagerBase(object):
             self.log.trace("Component info: {}".format(metadata))
             # Convert all values to str
             return dict([a, str(x)] for a, x in metadata.items())
-        # else:
+
         self.log.trace("Component not found in updateable components: {}"
                        .format(component_name))
         return {}
@@ -792,7 +1030,10 @@ class PeriphManagerBase(object):
 
         All key/value pairs are string -> string
         """
-        return {k: str(v) for k, v in iteritems(self._eeprom_head)}
+        return {
+            k: v.decode() if isinstance(v, bytes) else str(v)
+            for k, v in iteritems(self._eeprom_head)
+        }
 
     def set_mb_eeprom(self, eeprom_vals):
         """
@@ -853,7 +1094,7 @@ class PeriphManagerBase(object):
         the keys returned from this function can be used with
         get_chdr_link_options().
         """
-        raise NotImplementedError("get_chdr_link_types() not implemented.")
+        return list(self._xport_mgrs.keys())
 
     def get_chdr_link_options(self, xport_type):
         """
@@ -869,7 +1110,15 @@ class PeriphManagerBase(object):
         - link_rate (bps of the link, e.g. 10e9 for 10GigE)
 
         """
-        raise NotImplementedError("get_chdr_link_options() not implemented.")
+        if xport_type not in self._xport_mgrs:
+            self.log.warning(
+                f"Can't get link options for unknown link type: `{xport_type}'.")
+            return []
+        if xport_type == "udp":
+            return self._xport_mgrs[xport_type].get_chdr_link_options(
+                self.mboard_info['rpc_connection'])
+        # else:
+        return self._xport_mgrs[xport_type].get_chdr_link_options()
 
     #######################################################################
     # Claimer API
@@ -977,3 +1226,92 @@ class PeriphManagerBase(object):
         assert src in self.get_gpio_srcs(bank), \
             "Invalid GPIO source: {}".format(src)
         raise NotImplementedError("set_gpio_src() not available on this device!")
+
+    #######################################################################
+    # Sync API
+    #######################################################################
+    def get_clock_source(self):
+        " Returns the currently selected clock source "
+        raise NotImplementedError("get_clock_source() not available on this device!")
+
+    def get_time_source(self):
+        " Returns the currently selected time source "
+        raise NotImplementedError("get_time_source() not available on this device!")
+
+    def get_sync_source(self):
+        """
+        Gets the current time and clock source
+        """
+        return {
+            "time_source": self.get_time_source(),
+            "clock_source": self.get_clock_source(),
+        }
+
+    def get_clock_sources(self):
+        """
+        Returns a list of valid clock sources. This is a list of strings.
+        """
+        self.log.warning("get_clock_sources() was not specified for this device!")
+        return []
+
+    def get_time_sources(self):
+        """
+        Returns a list of valid time sources. This is a list of strings.
+        """
+        self.log.warning("get_time_sources() was not specified for this device!")
+        return []
+
+    def get_sync_sources(self):
+        """
+        Returns a list of valid sync sources. This is a list of dictionaries.
+        """
+        self.log.warning("get_sync_sources() was not specified for this device!")
+        return []
+
+    def set_clock_source(self, *args):
+        """
+        Set a clock source.
+
+        The choice to allow arbitrary arguments is based on historical decisions
+        and backward compatibility. UHD/mpmd will call this with a single argument,
+        so args[0] is the clock source (as a string).
+        """
+        raise NotImplementedError("set_clock_source() not available on this device!")
+
+    def set_time_source(self, time_source):
+        " Set a time source "
+        raise NotImplementedError("set_time_source() not available on this device!")
+
+    def set_sync_source(self, sync_args):
+        """
+        If a device has no special code for setting the sync-source atomically,
+        we simply forward these settings to set_clock_source() and set_time_source()
+        (in that order).
+        """
+        if sync_args not in self.get_sync_sources():
+            sync_args_str = \
+                ','.join([str(k) + '=' + str(v) for k, v in sync_args.items()])
+            self.log.warn(
+                f"Attempting to set unrecognized Sync source `{sync_args_str}'!")
+        clock_source = sync_args.get('clock_source', self.get_clock_source())
+        time_source = sync_args.get('time_source', self.get_time_source())
+        self.set_clock_source(clock_source)
+        self.set_time_source(time_source)
+
+    ###########################################################################
+    # Clock/Time API
+    ###########################################################################
+    def set_clock_source_out(self, enable=True):
+        """
+        Allows routing the clock configured as source to the RefOut terminal.
+        """
+        raise NotImplementedError("set_clock_source_out() not implemented.")
+
+    #######################################################################
+    # Discoverable Features
+    #######################################################################
+    def supports_feature(self, query):
+        """
+        Returns true if the queried feature is supported by a device.
+        """
+        return query in self.discoverable_features
